@@ -193,7 +193,6 @@ def duplicate_elimination(duplicatehandling_config):
 
 xp_persist_duration = 'child::cppa:PersistDuration/text()'
 def persistduration(duplicatehandling_config):
-    #logging.error(etree.tostring(duplicatehandling_config))
     return return_if_present_else_none(duplicatehandling_config, xp_persist_duration)
 
 xp_message_retry = 'child::cppa:RetryHandling'
@@ -289,10 +288,8 @@ def return_if_present_else_none(element, xpathexpr, default=None):
             logging.debug('No {}, returning default {}'.format(xpathexpr,default))
             return default
         else:
-            #logging.debug('No {} and no default, returning None'.format(xpathexpr))
             return None
     else:
-        #logging.debug('Found {} returning content'.format(xpathexpr))
         return exprL[0]
 
 def channel_transport_address(cpa, chid):
@@ -430,7 +427,11 @@ pmode_json_schema = {
                                               "properties" : {
                                                   "Algorithm" : {
                                                       "type" : "string"
+                                                  },
+                                                  "Certificate" : {
+                                                      "type" : "array"
                                                   }
+
                                               },
                                               "additionalProperties" : False
                                           },
@@ -888,10 +889,14 @@ def process_payloadservice(cpa, pmode, leg, channel):
         swapackage_config = swapackage(cpa, package_id)
         if swapackage_config != None:
             as4compressedparts = as4compressedmimepart(swapackage_config)
-            if len(as4compressedparts)>0:
+            if as4compressedparts != None:
                 pmode[leg]['PayloadService'] = {
                     'CompressionType' : 'application/gzip'
                 }
+            else:
+                logging.error(package_id+' '+str(len(as4compressedparts))+str(swapackage_config))
+
+
 
 def process_reception_awareness(cpa, pmode, leg, channel):
     receptionawareness_config = receptionawareness(channel)
@@ -969,8 +974,9 @@ def process_security(cpa, pmode, leg, channel):
         process_usernametoken(cpa, pmode, leg, security_config)
 
 def retrieve_certificate(cpa, certid):
-    return cpa.xpath('//cppa:Certificate[@id="{}"]//ds:X509Certificate/text()'.format(certid),
+    certtext = cpa.xpath('//cppa:Certificate[@id="{}"]//ds:X509Certificate/text()'.format(certid),
                      namespaces=NSMAP)
+    return certtext
 
 def process_signature(cpa, pmode, leg, security_config):
     xmlsignature = return_if_present_else_none(security_config,xp_signature)
