@@ -30,6 +30,10 @@ class CPPA3TestCase( unittest.TestCase ):
             # Else no schema validation
             self.schema = None
 
+        # If you set the following to True, always generate CPAs
+        # If False, preserve and check with previously created CPA
+        self.replacemode = False
+
     def do_unification(self, id):
         logging.info('------------------------------------------')
         logging.info('Running test {}'.format(id))
@@ -41,17 +45,23 @@ class CPPA3TestCase( unittest.TestCase ):
         bcpp =  lxml.etree.parse(cppb_file)
         if self.schema != None:
             if self.schema.validate(acpp) and self.schema.validate(bcpp):
-                pass
+                logging.info('CPP A and B are valid for {}'.format(id))
             else:
                 err = str(self.schema.error_log.last_error)
                 raise cppa3.unify.UnificationException('Invalid input: {}'.format(err))
         cpa = unify(acpp, bcpp)
-        if self.schema != None and self.schema.validate(cpa):
-            logging.error('CPA {} is invalid'.format(id))
-            raise cppa3.unify.UnificationException('CPA {} is invalid'.format(id))
+        if self.schema is not None:
+            if not self.schema.validate(cpa):
+                err = str(self.schema.error_log.last_error)
+                logging.error('CPA {} is invalid: {}'.format(id, err))
+                raise cppa3.unify.UnificationException('CPA {} is invalid: {}'.format(id, err))
+            else:
+                logging.info('Generated CPA valid for {}'.format(id))
+        else:
+            logging.info('CPA not validated for {}'.format(id))
 
         cpa_file_to_test_against = os.path.join(self.testdatadir,'cppa_ab_'+id+'.xml')
-        if os.path.isfile(cpa_file_to_test_against):
+        if not self.replacemode and os.path.isfile(cpa_file_to_test_against):
             base_cpa =  lxml.etree.parse(cpa_file_to_test_against)
             base_cpa_clean = reset_timestamped_values_for_compare(base_cpa)
             base_cpa_clean_string = lxml.etree.tostring(base_cpa_clean,
@@ -82,12 +92,7 @@ class CPPA3TestCase( unittest.TestCase ):
     def test_0001(self):
         self.do_unification('0001')
 
-    #def test_000X(self):
-    #    self.do_unification('0002')
-
-
     # One Service with with complementary Roles one non-matching Action, failure
-
     def test_0002(self):
         self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0002')
 
@@ -107,7 +112,7 @@ class CPPA3TestCase( unittest.TestCase ):
     def test_0009(self):
         self.do_unification('0009')
 
-    #Sample NamedProtocol for e-SENS
+    #Sample NamedProtocol for ENTSOG
     def test_0010(self):
         self.do_unification('0010')
 
@@ -119,7 +124,6 @@ class CPPA3TestCase( unittest.TestCase ):
     def test_0012(self):
         self.assertRaises( cppa3.unify.UnificationException, self.do_unification, '0012' )
 
-
     # A has two ServiceSpecification, only one of which is relevant for B
     def test_0013(self):
         self.do_unification('0013')
@@ -150,134 +154,29 @@ class CPPA3TestCase( unittest.TestCase ):
 
     # Some extra tests
     # Multiple ServiceSpecifications
-    def test_0200(self):
-        self.do_unification('0200')
-
-    def test_0201(self):
-        self.do_unification('0201')
-
-    def test_0202(self):
-        self.do_unification('0202')
-
-    def test_0203(self):
-        self.do_unification('0203')
-
-    # Extra test, failing ActivationDate
-    def test_0210(self):
-        self.assertRaises( cppa3.unify.UnificationException, self.do_unification, '0210' )
-
-    def test_0211(self):
-        self.assertRaises( cppa3.unify.UnificationException, self.do_unification, '0211' )
-
-
-
-    """
-    Certificates and Trust Anchors for Signing, Encryption and TLS
-    """
-
-    # Certificate and Trust Anchor test for Signing cert - positive - Send.
     def test_0020(self):
         self.do_unification('0020')
 
-    # Certificate and Trust Anchor test for Signing cert - negative -Send .
     def test_0021(self):
-        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0021')
+        self.do_unification('0021')
 
-    # Certificate and Trust Anchor test for Encryption cert - positive -Send.
     def test_0022(self):
         self.do_unification('0022')
 
-    # Certificate and Trust Anchor test for Encryption cert - negative.
     def test_0023(self):
-        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0023')
+        self.do_unification('0023')
 
-    # Certificate and Trust Anchor test for Signing cert - positive - Receive.
-    def test_0024(self):
-        self.do_unification('0024')
-
-    # Certificate and Trust Anchor test for Signing cert - negative - Receive.
-    def test_0025(self):
-        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0025')
-
-
-    # Certificate and Trust Anchor test for Encryption cert - positive -Receive.
-    def test_0026(self):
-        self.do_unification('0026')
-
-    # Certificate and Trust Anchor test for TLS Server cert - positive - Send.
-    def test_0027(self):
-        self.do_unification('0027')
-
-    # Certificate and Trust Anchor test for TLS Server cert - negative - Send.
-    def test_0028(self):
-        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0028')
-
-    # Certificate and Trust Anchor test for TLS Server cert - negative - Receive.
-    def test_0029(self):
-        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0029')
-
-    # Certificate and Trust Anchor test for TLS Server cert - positive - Receive.
-    def test_0030(self):
-        self.do_unification('0030')
-
-    # Certificate and Trust Anchor test for TLS Client cert - positive - Send.
-    def test_0031(self):
-        self.do_unification('0031')
-
-    # Certificate and Trust Anchor test for TLS Client cert - negative - Send.
-    def test_0032(self):
-        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0032')
-
-    # Certificate and Trust Anchor test for TLS Client cert - positive - Receive.
-    def test_0033(self):
-        self.do_unification('0033')
-
-    # Certificate and Trust Anchor test for TLS Client cert - negative - Receive.
-    def test_0034(self):
-        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0034')
-
-    """
-    TLS tests
-    """
-
-    # Cipher specified on one side
-    def test_0040(self):
-        self.do_unification('0040')
-
-    # Cipher specified on other side
-    def test_0041(self):
-        self.do_unification('0041')
-
-    # Cipher specified on both sides,  intersection exists
-    def test_0042(self):
-        self.do_unification('0042')
-
-    # Cipher specified on both sides,  intersection does not exist
-    def test_0043(self):
-        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0043')
-
-    # Version mismatch
-
-    """
-    User Authentication
-    """
-
-    # Specified on both sides consistently:  success
+    # Extra test, failing ActivationDate
     def test_0050(self):
-        self.do_unification('0050')
+        self.assertRaises( cppa3.unify.UnificationException, self.do_unification, '0210' )
 
-    # Specified on both sides inconsistently:  failure
     def test_0051(self):
-        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0051')
-
-    # Specified on both sides inconsistently:  failure
-    def test_0052(self):
-        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0052')
-
-    # Specified on both sides inconsistently:  failure
-    def test_0053(self):
-        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0053')
+        self.assertRaises( cppa3.unify.UnificationException, self.do_unification, '0211' )
 
+    # Named Channel variant of 1007
+    @unittest.skip
+    def test_0060(self):
+        self.do_unification('0060')
 
     """
     Payload Profile
@@ -288,7 +187,6 @@ class CPPA3TestCase( unittest.TestCase ):
     def test_0100(self):
         self.do_unification('0100')
 
-
     # Single part, non-matching partname, matching namespace, matching root element,
     # matching location, matching cardinality
     def test_0101(self):
@@ -323,6 +221,123 @@ class CPPA3TestCase( unittest.TestCase ):
     def test_0107(self):
         self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0107')
 
+    """
+    0200 ...  reserved for future use ...
+    """
+
+
+    """
+    0300 Certificates and Trust Anchors for Signing, Encryption and TLS
+    """
+
+    # Certificate and Trust Anchor test for Signing cert - positive - Send.
+    def test_0300(self):
+        self.do_unification('0300')
+
+    # Certificate and Trust Anchor test for Signing cert - negative -Send .
+    def test_0301(self):
+        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0301')
+
+    # Certificate and Trust Anchor test for Encryption cert - positive -Send.
+    def test_0302(self):
+        self.do_unification('0302')
+
+    # Certificate and Trust Anchor test for Encryption cert - negative.
+    def test_0303(self):
+        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0303')
+
+    # Certificate and Trust Anchor test for Signing cert - positive - Receive.
+    def test_0304(self):
+        self.do_unification('0304')
+
+    # Certificate and Trust Anchor test for Signing cert - negative - Receive.
+    def test_0305(self):
+        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0305')
+
+    # Certificate and Trust Anchor test for Encryption cert - positive -Receive.
+    def test_0306(self):
+        self.do_unification('0306')
+
+    # Certificate and Trust Anchor test for TLS Server cert - positive - Send.
+    def test_0307(self):
+        self.do_unification('0307')
+
+    # Certificate and Trust Anchor test for TLS Server cert - negative - Send.
+    def test_0308(self):
+        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0308')
+
+    # Certificate and Trust Anchor test for TLS Server cert - negative - Receive.
+    def test_0309(self):
+        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0309')
+
+    # Certificate and Trust Anchor test for TLS Server cert - positive - Receive.
+    def test_0310(self):
+        self.do_unification('0310')
+
+    # Certificate and Trust Anchor test for TLS Client cert - positive - Send.
+    def test_0311(self):
+        self.do_unification('0311')
+
+    # Certificate and Trust Anchor test for TLS Client cert - negative - Send.
+    def test_0312(self):
+        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0312')
+
+    # Certificate and Trust Anchor test for TLS Client cert - positive - Receive.
+    def test_0313(self):
+        self.do_unification('0313')
+
+    # Certificate and Trust Anchor test for TLS Client cert - negative - Receive.
+    def test_0314(self):
+        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0314')
+
+
+    """
+    0400 TLS tests
+    """
+
+    # Cipher specified on one side
+    def test_0400(self):
+        self.do_unification('0400')
+
+    # Cipher specified on other side
+    def test_0401(self):
+        self.do_unification('0401')
+
+    # Cipher specified on both sides,  intersection exists
+    def test_0402(self):
+        self.do_unification('0402')
+
+    # Cipher specified on both sides,  intersection does not exist
+    def test_0403(self):
+        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0403')
+
+    # Version mismatch
+    def test_0404(self):
+        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0404')
+
+
+    """
+    0500
+    User Authentication
+    """
+
+    # Specified on both sides consistently:  success
+    def test_0500(self):
+        self.do_unification('0500')
+
+    # Specified on both sides inconsistently:  failure
+    def test_0501(self):
+        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0501')
+
+    # Specified on both sides inconsistently:  failure
+    def test_0502(self):
+        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0502')
+
+    # Specified on both sides inconsistently:  failure
+    def test_0503(self):
+        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '0503')
+
+
 
     """
     ebBP attributes test
@@ -330,6 +345,9 @@ class CPPA3TestCase( unittest.TestCase ):
     @@@ to do
     """
 
+
+
+
     """
     ebMS3 tests
     """
@@ -348,6 +366,11 @@ class CPPA3TestCase( unittest.TestCase ):
         # two delivery channels,  the first one is non compatible so the second must be chosen
         self.do_unification('1006')
 
+    @unittest.skip
+    def test_1007(self):
+        # example involving wider variety of services
+        self.do_unification('1007')
+
     def test_1010(self):
         # E-SENS AS4 PoC
         self.do_unification('1010')
@@ -451,8 +474,6 @@ class CPPA3TestCase( unittest.TestCase ):
     def test_1145(self):
         self.do_unification('1145')
 
-    # Error:  conflict in specified MPC,  gevallen A t/m F
-
     # Simple AS4 Push MEP Send with a single Payload part and a Compressed Payload
     def test_1146(self):
         self.do_unification('1146')
@@ -461,6 +482,19 @@ class CPPA3TestCase( unittest.TestCase ):
     def test_1147(self):
         self.do_unification('1147')
 
+    # Various MPC errors
+    # Both Sender and Receiver set the MPC,  but incompatibly
+    def test_1150(self):
+        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '1150')
+
+    # Pull client cannot set the MPC for the server, A sends
+    def test_1151(self):
+        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '1151')
+
+    # Pull client cannot set the MPC for the server, A receives
+    def test_1152(self):
+        self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '1152')
+
 
     # 1200 range
     # Reliable Messaging:  AS4 reception awareness
@@ -505,16 +539,10 @@ class CPPA3TestCase( unittest.TestCase ):
     def test_1400(self):
         self.do_unification('1400')
 
-
     # External Payload
     def test_1410(self):
         self.do_unification('1410')
 
-    # Extensibility
-
-    # 1500
-    # Cartesian
-
 
 
     """
@@ -639,8 +667,6 @@ class CPPA3TestCase( unittest.TestCase ):
     def test_2149(self):
         self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '2149')
 
-
-
     # User Authentication
 
     # Specified on both sides consistently:  success
@@ -651,8 +677,6 @@ class CPPA3TestCase( unittest.TestCase ):
     #def test_2170(self):
     #    self.assertRaises(cppa3.unify.UnificationException, self.do_unification, '2171')
 
-
-
     # Packaging
 
 
@@ -722,8 +746,6 @@ class CPPA3TestCase( unittest.TestCase ):
     def test_4031(self):
         self.do_unification('4031')
 
-    def test_09900(self):
-        self.do_unification('0100')
 
 
     # Synchronous Business Response
@@ -740,10 +762,202 @@ class CPPA3TestCase( unittest.TestCase ):
     Extensibility
 
     """
-    def test_0100(self):
-        self.do_unification('0100')
+    #def test_0100(self):
+    #    self.do_unification('0100')
 
+    """
+    Tests for lower level functions
 
+    unify_simple_subelement
+    """
+
+    def test_9000(self):
+        logging.info('Running test 9000')
+        unifier = cppa3.unify.CPABuilder(nsmap={'ns':'urn:namespace',
+                                                'ns2':'urn:namespace2'})
+
+        a_element = lxml.etree.fromstring("""<ns:a xmlns:ns="urn:namespace" xmlns:ns2="urn:namespace2">
+        </ns:a>
+        """)
+        b_element = lxml.etree.fromstring("""<ns:a xmlns:ns="urn:namespace" xmlns:ns2="urn:namespace2">
+        </ns:a>
+        """)
+        ab_element = lxml.etree.fromstring("""
+        <ns:a xmlns:ns="urn:namespace" xmlns:ns2="urn:namespace2"/>
+        """)
+        unifier.unify_simple_subelement(a_element,
+                                        b_element,
+                                        ab_element,
+                                        'ns2',
+                                        'b',
+                                        intersectifmultiple=True,
+                                        strictelements=False,
+                                        required=False)
+        logging.debug(lxml.etree.tostring(ab_element, pretty_print=True))
+
+    def _test_9001(self):
+        logging.info('Running test 9001')
+        unifier = cppa3.unify.CPABuilder(nsmap={'ns':'urn:namespace',
+                                                'ns2':'urn:namespace2'})
+
+        a_element = lxml.etree.fromstring("""<ns:a xmlns:ns="urn:namespace" xmlns:ns2="urn:namespace2">
+        </ns:a>
+        """)
+        b_element = lxml.etree.fromstring("""<ns:a xmlns:ns="urn:namespace" xmlns:ns2="urn:namespace2">
+        </ns:a>
+        """)
+        ab_element = lxml.etree.fromstring("""
+        <ns:a xmlns:ns="urn:namespace" xmlns:ns2="urn:namespace2"/>
+        """)
+        unifier.unify_simple_subelement(a_element,
+                                        b_element,
+                                        ab_element,
+                                        'ns2',
+                                        'b',
+                                        intersectifmultiple=True,
+                                        strictelements=False,
+                                        required=True)
+
+    def test_9001(self):
+        self.assertRaises(cppa3.unify.UnificationException, self._test_9001)
+
+    def test_9002(self):
+        logging.info('Running test 9002')
+        unifier = cppa3.unify.CPABuilder(nsmap={'ns':'urn:namespace',
+                                                'ns2':'urn:namespace2'})
+
+        a_element = lxml.etree.fromstring("""<ns:a xmlns:ns="urn:namespace" xmlns:ns2="urn:namespace2">
+        <ns2:b>value1</ns2:b>
+        </ns:a>
+        """)
+        b_element = lxml.etree.fromstring("""<ns:a xmlns:ns="urn:namespace" xmlns:ns2="urn:namespace2">
+        </ns:a>
+        """)
+        ab_element = lxml.etree.fromstring("""
+        <ns:a xmlns:ns="urn:namespace" xmlns:ns2="urn:namespace2"/>
+        """)
+        unifier.unify_simple_subelement(a_element,
+                                        b_element,
+                                        ab_element,
+                                        'ns2',
+                                        'b',
+                                        intersectifmultiple=True,
+                                        strictelements=False,
+                                        required=True)
+        logging.debug(lxml.etree.tostring(ab_element, pretty_print=True))
+
+    def test_9003(self):
+        logging.info('Running test 9003')
+        unifier = cppa3.unify.CPABuilder(nsmap={'ns':'urn:namespace',
+                                                'ns2':'urn:namespace2'})
+
+        a_element = lxml.etree.fromstring("""<ns:a xmlns:ns="urn:namespace" xmlns:ns2="urn:namespace2">
+        </ns:a>
+        """)
+        b_element = lxml.etree.fromstring("""<ns:a xmlns:ns="urn:namespace" xmlns:ns2="urn:namespace2">
+        <ns2:b>value1</ns2:b>
+        </ns:a>
+        """)
+        ab_element = lxml.etree.fromstring("""
+        <ns:a xmlns:ns="urn:namespace" xmlns:ns2="urn:namespace2"/>
+        """)
+        unifier.unify_simple_subelement(a_element,
+                                        b_element,
+                                        ab_element,
+                                        'ns2',
+                                        'b',
+                                        intersectifmultiple=True,
+                                        strictelements=False,
+                                        required=True)
+        logging.debug(lxml.etree.tostring(ab_element, pretty_print=True))
+
+    def test_9004(self):
+        logging.info('Running test 9004')
+        unifier = cppa3.unify.CPABuilder(nsmap={'ns':'urn:namespace',
+                                                'ns2':'urn:namespace2'})
+
+        a_element = lxml.etree.fromstring("""<ns:a xmlns:ns="urn:namespace" xmlns:ns2="urn:namespace2">
+        <ns2:b>value1</ns2:b>
+        <ns2:b>value2</ns2:b>
+        <ns2:b>value3</ns2:b>
+        </ns:a>
+        """)
+        b_element = lxml.etree.fromstring("""<ns:a xmlns:ns="urn:namespace" xmlns:ns2="urn:namespace2">
+        <ns2:b>value2</ns2:b>
+        <ns2:b>value3</ns2:b>
+        <ns2:b>value4</ns2:b>
+        </ns:a>
+        """)
+        ab_element = lxml.etree.fromstring("""
+        <ns:a xmlns:ns="urn:namespace" xmlns:ns2="urn:namespace2"/>
+        """)
+        unifier.unify_simple_subelement(a_element,
+                                        b_element,
+                                        ab_element,
+                                        'ns2',
+                                        'b',
+                                        intersectifmultiple=True,
+                                        strictelements=False)
+        logging.debug(lxml.etree.tostring(ab_element, pretty_print=True))
+
+    def test_9005(self):
+        logging.info('Running test 9005')
+        unifier = cppa3.unify.CPABuilder(nsmap={'ns':'urn:namespace',
+                                                'ns2':'urn:namespace2'})
+
+        a_element = lxml.etree.fromstring("""<ns:a xmlns:ns="urn:namespace" xmlns:ns2="urn:namespace2">
+        <ns2:b>value1</ns2:b>
+        <ns2:b>value2</ns2:b>
+        <ns2:b>value3</ns2:b>
+        </ns:a>
+        """)
+        b_element = lxml.etree.fromstring("""<ns:a xmlns:ns="urn:namespace" xmlns:ns2="urn:namespace2">
+        <ns2:b>value2</ns2:b>
+        <ns2:b>value3</ns2:b>
+        <ns2:b>value4</ns2:b>
+        </ns:a>
+        """)
+        ab_element = lxml.etree.fromstring("""
+        <ns:a xmlns:ns="urn:namespace" xmlns:ns2="urn:namespace2"/>
+        """)
+        unifier.unify_simple_subelement(a_element,
+                                        b_element,
+                                        ab_element,
+                                        'ns2',
+                                        'b',
+                                        intersectifmultiple=False,
+                                        strictelements=False)
+        logging.debug(lxml.etree.tostring(ab_element, pretty_print=True))
+
+    def _test_9006(self):
+        logging.info('Running test 9006')
+        unifier = cppa3.unify.CPABuilder(nsmap={'ns':'urn:namespace',
+                                                'ns2':'urn:namespace2'})
+
+        a_element = lxml.etree.fromstring("""<ns:a xmlns:ns="urn:namespace" xmlns:ns2="urn:namespace2">
+        <ns2:b>value1</ns2:b>
+        <ns2:b>value2</ns2:b>
+        </ns:a>
+        """)
+        b_element = lxml.etree.fromstring("""<ns:a xmlns:ns="urn:namespace" xmlns:ns2="urn:namespace2">
+        <ns2:b>value3</ns2:b>
+        <ns2:b>value4</ns2:b>
+        </ns:a>
+        """)
+        ab_element = lxml.etree.fromstring("""
+        <ns:a xmlns:ns="urn:namespace" xmlns:ns2="urn:namespace2"/>
+        """)
+        unifier.unify_simple_subelement(a_element,
+                                        b_element,
+                                        ab_element,
+                                        'ns2',
+                                        'b',
+                                        intersectifmultiple=True,
+                                        strictelements=False,
+                                        required=False)
+
+    def test_9006(self):
+        self.assertRaises(cppa3.unify.UnificationException, self._test_9006)
 
 def reset_timestamped_values_for_compare(cpa):
     cpa_to_edit = deepcopy(cpa)
@@ -764,3 +978,5 @@ def reset_timestamped_values_for_compare(cpa):
 
 
 
+
+
